/*
**  CXSC is a C++ library for eXtended Scientific Computing (V 2.0)
**
**  Copyright (C) 1990-2000 Institut fuer Angewandte Mathematik,
**                          Universitaet Karlsruhe, Germany
**            (C) 2000-2004 Wiss. Rechnen/Softwaretechnologie
**                          Universitaet Wuppertal, Germany   
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Library General Public
**  License as published by the Free Software Foundation; either
**  version 2 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Library General Public License for more details.
**
**  You should have received a copy of the GNU Library General Public
**  License along with this library; if not, write to the Free
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
/* CVS $Id: l_comple.hpp,v 1.5 2004/03/11 17:49:06 cxsc Exp $ */

#ifndef L_COMPLEX_HPP
#define L_COMPLEX_HPP

#include <iostream>
#include <string>

#include <except.hpp>
#include <complex.hpp>
#include <cdot.hpp>
#include <l_real.hpp>
#include <l_rmath.hpp>

namespace cxsc {

class l_complex
{
   private:
    // ------------- Datenelemente -------------------------------------------
    l_real re, im;

   public:
    // ------------- Constructors --------------------------------------------
    l_complex(void)  throw() {}
    l_complex(const l_real& a, const l_real& b) throw() : re(a), im(b) { }
    l_complex(const real& a, const real& b) throw() : re(a), im(b) { }

    l_complex & operator = (const l_real & lr) throw() 
        { re = lr; im = 0; return *this; }
    l_complex & operator = (const real & r) throw() 
      { re = r; im = 0; return *this; } 
    l_complex & operator = (const complex & c) throw() 
        { re = Re(c); im = Im(c); return *this; } 
    l_complex & operator = (const dotprecision & d) throw() 
        { re = d; im = 0.0; return *this; }
    l_complex & operator = (const cdotprecision & cd) throw() 
        { re = Re(cd); im = Im(cd); return *this; }

    // ------------- Type-Casts ----------------------------------------------
    explicit inline l_complex(const l_real  &r) throw() : re(r), im(0.0) { }
    explicit inline l_complex(const   real  &r) throw() : re(r), im(0.0) { }
    explicit inline l_complex(const complex &r) throw() : re(Re(r)), im(Im(r))
                                                                         { }
    explicit inline l_complex(const dotprecision &d) throw()
                  : re(d), im(0.0) { }
    explicit inline l_complex(const cdotprecision &cd) throw()
	          : re(Re(cd)), im(Im(cd)) { }

    friend inline l_complex _l_complex(const l_real &a) throw()
	{ return l_complex(a); }
    friend inline l_complex _l_complex(const l_real &a, const l_real &b)
        throw()   { return l_complex(a,b); }
    friend inline l_complex _l_complex(const real &a) throw()
	{ return l_complex(a); }
    friend inline l_complex _l_complex(const real &a, const real &b)
        throw()   { return l_complex(a,b); }
    friend inline l_complex _l_complex(const complex &c)
        throw()   { return l_complex(c); }
    friend inline l_complex _l_complex(const dotprecision &d)
        throw()   { return l_complex(d); }
    friend inline l_complex _l_complex(const cdotprecision &cd)
        throw()   { return l_complex(cd); }

    // ----------------------------------------------------------------------
    friend int StagPrec(const l_complex&) throw();

// ------------- Arithmetic Operators ---------------------------------------
    // ----------------------------------------------------------------------
    // ------------- Unary Operators ----------------------------------------

    friend l_complex operator-(const l_complex&);
    friend l_complex operator+(const l_complex&);

    // ------------- Binary Operators ---------------------------------------
    // ----------------- l_complex +/- l_complex ----------------------------
    friend inline l_complex operator +
                   (const l_complex &, const l_complex &) throw();

    friend inline l_complex operator -
                   (const l_complex &, const l_complex &) throw();
    
    // ----------------- l_complex + complex --------------------------------
    inline l_complex operator + (const complex &) const throw();

    friend inline l_complex operator +
                   (const complex &, const l_complex &) throw();

    // ---------------- l_complex + real ------------------------------------
    inline l_complex operator + (const real &) const throw();

    friend inline l_complex operator +
                   (const real &, const l_complex &) throw();

    // ---------------- l_complex + l_real ----------------------------------
    inline l_complex operator + (const l_real &) const throw();

    friend inline l_complex operator +
                   (const l_real &, const l_complex &) throw();

    // ---------------- l_complex - l_real ----------------------------------
    inline l_complex operator - (const l_real &) const throw();

    friend inline l_complex operator -
                   (const l_real &, const l_complex &) throw();

    // ----------------- l_complex - complex -------------------------------
    inline l_complex operator - (const complex &) const throw();

    friend inline l_complex operator -
                   (const complex &, const l_complex &) throw();

    // ---------------- l_complex - real ------------------------------------
    inline l_complex operator - (const real &) const throw();

    friend inline l_complex operator -
                   (const real &, const l_complex &) throw();

    // ---------------- l_complex + cdotprecision ---------------------------
    friend cdotprecision operator+ 
      (const l_complex &, const cdotprecision &) throw();

    friend cdotprecision operator+ 
      (const cdotprecision &, const l_complex &) throw();

    // ---------------- l_complex - cdotprecision ---------------------------
    friend cdotprecision operator- 
      (const l_complex &, const cdotprecision &) throw();

    friend cdotprecision operator- 
      (const cdotprecision &, const l_complex &) throw(); 

    // ---------------- l_complex + dotprecision ----------------------------
    friend cdotprecision operator+ 
      (const l_complex &, const dotprecision &) throw();

    friend cdotprecision operator+ 
      (const dotprecision &, const l_complex &) throw();

    // ---------------- l_complex - dotprecision ----------------------------
    friend cdotprecision operator- 
      (const l_complex &, const dotprecision &) throw();

    friend cdotprecision operator- 
      (const dotprecision &, const l_complex &) throw();


// ------------- Multiplication ---------------------------------------------
// ------------------ l_complex * l_complex ---------------------------------

friend l_complex operator * (const l_complex& a, const l_complex& b)
    throw();

// ------------------ l_complex * complex -----------------------------------
friend l_complex operator * (const l_complex& a, const complex& b)
    throw();

friend l_complex operator * ( const complex& b, const l_complex& a )
    throw();

// ------------------ l_complex * real ---------------------------------------
    inline l_complex operator * (const real &) const throw();

    friend inline l_complex operator *
                   (const real &, const l_complex &) throw();

// ------------------ l_complex * l_real -------------------------------------
    inline l_complex operator * (const l_real &) const throw();

    friend inline l_complex operator *
                   (const l_real &, const l_complex &) throw();


// ----------------- Others --------------------------------------------------
    friend l_real & Re(l_complex& a) { return a.re; }
    friend l_real   Re(const l_complex& a) { return a.re; }
    friend l_real & Im(l_complex& a) { return a.im; }
    friend l_real   Im(const l_complex& a) { return a.im; }

    l_complex conj(const l_complex& lc) // returning the conjugated value
                        { return l_complex(lc.re,-lc.im); }
    friend l_complex & SetRe(l_complex & a,const l_real & b) 
	{ a.re=b; return a; } // The real part of a is substituted by b.
    // SetRe(lc,lr); --> Re(lc)=lr;
    // lc1 = SetRe(lc,lr); --> Re(lc)=lr; and lc1 = lc;
    friend l_complex & SetIm(l_complex & a,const l_real & b) 
        { a.im=b; return a; } // See SetRe(...);


// ----- accumulate(cdotprecision,l_complex,l_complex|complex|real|l_real ----
    friend void accumulate(cdotprecision&, const l_complex&, 
                                           const l_complex&) throw();

    friend void accumulate(cdotprecision&, const l_complex&, 
                                           const complex&) throw();

    friend void accumulate(cdotprecision&, const l_complex&, 
                                           const real&) throw();

    friend void accumulate(cdotprecision&, const l_complex&, 
                                           const l_real&) throw();

// ---------------- cdotprecision +(-)= l_complex ---------------------------- 
friend inline cdotprecision & operator += (cdotprecision &cd, 
                                           const l_complex &lc) throw() 
	{
	    cd = cd + lc; return cd;
	}
friend inline cdotprecision & operator -= (cdotprecision &cd, 
                                           const l_complex &lc) throw() 
	{
	    cd = cd - lc; return cd;
	}

// ---------------- l_complex +(-)= l_complex|complex|real|l_real ------------
friend inline l_complex & operator += (l_complex &,const l_complex &) throw(); 
friend inline l_complex & operator -= (l_complex &,const l_complex &) throw(); 
friend inline l_complex & operator += (l_complex &,const complex &) throw(); 
friend inline l_complex & operator -= (l_complex &,const complex &) throw();
friend inline l_complex & operator += (l_complex &,const real &) throw();
friend inline l_complex & operator -= (l_complex &,const real &) throw();
friend inline l_complex & operator += (l_complex &,const l_real &) throw();
friend inline l_complex & operator -= (l_complex &,const l_real &) throw(); 

// ---------------- l_complex *= l_complex|complex|real|l_real ---------------
friend inline l_complex & operator *= (l_complex &,const l_complex &) throw();
friend inline l_complex & operator *= (l_complex &,const complex &) throw();
friend inline l_complex & operator *= (l_complex &,const real &) throw();
friend inline l_complex & operator *= (l_complex &,const l_real &) throw();

// ---------------- Compare Operators ----------------------------------------
friend inline bool operator! (const l_complex &) throw();
friend inline bool operator== (const l_complex &, const l_complex &) throw();
friend inline bool operator!= (const l_complex &, const l_complex &) throw();
friend inline bool operator== (const l_complex &, const complex &) throw();
friend inline bool operator== (const complex &, const l_complex &) throw();
friend inline bool operator!= (const l_complex &, const complex &) throw();
friend inline bool operator!= (const complex &, const l_complex &) throw();
friend inline bool operator== (const l_complex &, const real &) throw();
friend inline bool operator== (const real &, const l_complex &) throw();
friend inline bool operator!= (const l_complex &, const real &) throw();
friend inline bool operator!= (const real &, const l_complex &) throw();
friend inline bool operator== (const l_complex &, const l_real &) throw();
friend inline bool operator== (const l_real &, const l_complex &) throw();
friend inline bool operator!= (const l_complex &, const l_real &) throw();
friend inline bool operator!= (const l_real &, const l_complex &) throw();
friend inline bool operator== (const l_complex &, const dotprecision &)
                                                                  throw();
friend inline bool operator== (const dotprecision &, const l_complex &)
                                                                  throw();
friend inline bool operator!= (const l_complex &, const dotprecision &)
                                                                  throw();
friend inline bool operator!= (const dotprecision &, const l_complex &)
                                                                  throw();

friend inline bool operator ==(const cdotprecision &, const l_complex &)
    throw(); // {l_complex.inl}
friend inline bool operator ==(const l_complex &, const cdotprecision &)
    throw(); // {l_complex.inl}
friend inline bool operator !=(const cdotprecision &, const l_complex &)
    throw(); // {l_complex.inl}
friend inline bool operator !=(const l_complex &, const cdotprecision &)
    throw(); // {l_complex.inl}

// -------------- Division: Directed rounding, Blomquist 28.11.02 ------------

    friend l_complex divn (const l_complex &, const l_complex &);
    friend l_complex divd (const l_complex &, const l_complex &);
    friend l_complex divu (const l_complex &, const l_complex &);

// -------------- Division:  Blomquist 28.11.02 ------------------------------
    friend l_complex operator / (const l_complex &,const l_complex &) throw();

    friend inline l_complex operator / (const l_complex & a,
                                        const complex & b) throw();
    friend inline l_complex operator / (const l_complex & a,
                                        const l_real & b) throw();
    friend inline l_complex operator / (const l_complex & a,
                                        const real & b) throw();

    friend inline l_complex operator / (const complex& a,
                                        const l_complex& b) throw();
    friend inline l_complex operator / (const real& a,
                                        const l_complex& b) throw();
    friend inline l_complex operator / (const l_real& a,
                                        const l_complex& b) throw();

    friend inline l_complex& operator /=(l_complex&,const l_complex&) throw();
    friend inline l_complex& operator /=(l_complex&,const complex&) throw();
    friend inline l_complex& operator /=(l_complex&,const real&) throw();
    friend inline l_complex& operator /=(l_complex&,const l_real&) throw();

friend l_real abs2(const l_complex &a) throw(); // a.re*a.re + a.im*a.im;
friend l_real abs (const l_complex &z) throw(); 

// ----------------------- Output --------------------------------------------

friend std::ostream& operator << (std::ostream& s,const l_complex& z ) throw()
// A complex number z of type l_complex is written to the output channel.
{     
    s << '('          
    << z.re << ", "   
    << z.im       
    << ')';
    return s;
}

friend std::string & operator << (std::string &s, const l_complex& a) throw()
// The value of a variable a of type l_complex is copied to a string s.
// s has the form:  (Re(a),Im(a))
{  
    s+='(';
    s << a.re;
    s+=", ";
    s << a.im; 
    s+=')';
    return s;
}

// ----------------------- Input ---------------------------------------------

friend std::istream & operator >> (std::istream &s, l_complex &a) throw()
// An input of a complex number z of the form (Re(z),Im(z)) is copied to
// the variable a of type l_complex.
{  
   char c;

   skipeolnflag = inpdotflag = true;
   c = skipwhitespacessinglechar (s, '(');
   if (inpdotflag) 
      s.putback(c);

   s >> a.re;

   skipeolnflag = inpdotflag = true;
   c = skipwhitespacessinglechar (s, ',');
   if (inpdotflag) s.putback(c);

   s >> a.im >> RestoreOpt;

   if (!waseolnflag) 
   {
      skipeolnflag = false, inpdotflag = true;
      c = skipwhitespaces (s);
      if (inpdotflag && c != ')') 
         s.putback(c);
   }
   return s;
}

friend std::string & operator >> (std::string &s, l_complex &a) throw()
// A complex number z of the form (Re(z),Im(z)), represented in a string s
// is copied to a of type l_complex.
{   
   s = skipwhitespacessinglechar (s, '(');
   s >> SaveOpt >> a.re;
   s = skipwhitespacessinglechar (s, ',');
   s >> a.im >> RestoreOpt;
   s = skipwhitespaces (s);

   if (s[0] == ')') 
      s.erase(0,1);
   return s;
}

}; // end of class l_complex

}  // end namespace cxsc

#include "l_comple.inl"

#endif














